class SiteContact:
    def __init__(self, **kwargs):
        self.id = kwargs.get("id")
        self.firstName = kwargs.get("firstName")
        self.surname = kwargs.get("surname")
        self.fullname = kwargs.get("fullname")



class Device:
    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.description = kwargs.get("description")
        self.primaryIP4Id = kwargs.get("primaryIP4Id")
        self.primaryIP4Address = kwargs.get("primaryIP4Address")
        self.serialNumber = kwargs.get("serialNumber")
        self.locationId = kwargs.get("locationId")
        self.locationName = kwargs.get("locationName")
        self.siteId = kwargs.get("siteId")
        self.siteName = kwargs.get("siteName")
        self.organizationId = kwargs.get("organizationId")
        self.organizationName = kwargs.get("organizationName")
        self.rackId = kwargs.get("rackId")
        self.rackFace = kwargs.get("rackFace")
        self.validFrom = kwargs.get("validFrom")
        self.validTo = kwargs.get("validTo")
        self.isActive = kwargs.get("isActive")
        self.id = kwargs.get("id")

        self.netBoxId = kwargs.get("netBoxId")

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name



from loguru import logger
from slugify import slugify

from komora_syncer.helpers.utils import build_tenant_name, clean_organization_name
from komora_syncer.models.netbox.NetboxBase import NetboxBase
from komora_syncer.models.netbox.NetboxCache import NetboxCache


class NbTenant(NetboxBase):
    def __init__(self, komora_obj=None):
        super().__init__()

        self.name = build_tenant_name(komora_obj.clientId, komora_obj.name)
        self.slug = slugify(self.name) if slugify(self.name) else komora_obj.id
        self.description = (
            " / ".join(komora_obj.keywords) if isinstance(komora_obj.keywords, list) else komora_obj.keywords
        )

        try:
            komora_vip_note = komora_obj.vipNote.strip()
        except AttributeError:
            komora_vip_note = ""

        self.custom_fields = {
            "client_id": komora_obj.clientId,
            "komora_id": komora_obj.id,
            "client_name": clean_organization_name(komora_obj.name),
            "komora_is_customer": komora_obj.isCustomer,
            "komora_is_supplier": komora_obj.isSupplier,
            "komora_is_vip": komora_obj.isVip,
            "komora_is_actual": komora_obj.isActual,
            "komora_is_member": komora_obj.isMember,
            "komora_vip_note": komora_vip_note,
        }

        self.api_object = None

    def find_or_create(self):
        self.find()
        if not self.api_object:
            self.create()

        return self.api_object

    def find(self):
        komora_id = self.custom_fields.get("komora_id")
        tenant_cache = NetboxCache.get_tenant_cache()

        try:
            if komora_id in tenant_cache:
                netbox_tenant = tenant_cache[komora_id]
            else:
                netbox_tenant = self.nb.tenancy.tenants.get(cf_komora_id=komora_id)

            if netbox_tenant:
                self.api_object = netbox_tenant
        except Exception:
            logger.exception(f"Unable to find tenant by komora_id: {komora_id}")
            netbox_tenant = None

        if not netbox_tenant:
            try:
                netbox_tenant = self.nb.tenancy.tenants.get(name__ie=self.name)

                if netbox_tenant:
                    logger.warning(f"komora_id: {str(komora_id)} was not found, but Tenant {self.name} already exists")
                    self.api_object = netbox_tenant
            except Exception:
                logger.exception(f"Unable to find tenant by name: {self.name}")
                raise

        return self.api_object

    def create(self):
        try:
            params = self.params
            netbox_tenant = self.nb.tenancy.tenants.create(params)

            logger.info(f"Region: {self.name} created successfully")
            self.api_object = netbox_tenant
        except Exception:
            logger.exception(f"Unable to create netbox tenant: {self.name}")
            raise

        return self.api_object

    def update(self, nb_tenant):
        try:
            # TODO: HotFix: Ensure slug is never empty
            if not self.params["slug"]:
                komora_id = self.params["custom_fields"].get("komora_id")
                if komora_id:
                    self.slug = komora_id
                else:
                    # Fallback if komora_id is also empty
                    self.slug = f"tenant-{self.name}"[:63]
            if nb_tenant.update(self.params):
                self.api_object = nb_tenant
                logger.info(f"Tenant: {self.name} updated successfully")
        except Exception:
            logger.exception(f"Unable to update tenant: {self.name} {nb_tenant.id}", exc_info=False)

    def synchronize(self):
        tenant = self.find()

        if tenant:
            self.update(tenant)
        else:
            self.create()

    @property
    def params(self):
        return {
            "name": self.name,
            "slug": self.slug,
            "description": self.description,
            "custom_fields": self.custom_fields,
        }



import os
import sys

import appdirs
import yaml
from loguru import logger

from komora_syncer import __appname__

os.environ["XDG_CONFIG_DIRS"] = "/etc"
CONFIG_DIRS = (
    appdirs.user_config_dir(__appname__),
    appdirs.site_config_dir(__appname__),
)
CONFIG_FILENAME = "config.yml"


def get_config():
    """
    Get config file and load it with yaml
    :returns: loaded config in yaml, as a dict object
    """

    if getattr(get_config, "cache", None):
        return get_config.cache

    if os.environ.get("CONFIG_FOLDER_PATH"):
        config_path = os.path.join(os.environ.get("CONFIG_FOLDER_PATH"), CONFIG_FILENAME)
    else:
        for d in CONFIG_DIRS:
            config_path = os.path.join(d, CONFIG_FILENAME)
            if os.path.isfile(config_path):
                break
        else:
            logger.error(
                "No configuration file can be found. Please create a "
                "config.yml in one of these directories:\n"
                "{}".format(", ".join(CONFIG_DIRS))
            )
            exit(0)

    try:
        with open(config_path, "r") as config_file:
            conf = yaml.safe_load(config_file)
            get_config.cache = conf
            return conf
    except FileNotFoundError as e:
        logger.error(e)
        exit(0)


def configure_logger():
    logging_config = get_config().get("logging", {})
    level = logging_config.get("LOG_LEVEL", "DEBUG")

    logger.remove()

    # Add console sink (stdout) with formatting and level filtering
    logger.add(sys.stdout, level=level)



class Location:
    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.description = kwargs.get("description")
        self.roomPin = kwargs.get("roomPin")
        self.parentId = kwargs.get("parentId")
        self.id = kwargs.get("id")

        if kwargs.get("locations"):
            self.locations = []
            for location in kwargs.get("locations"):
                self.locations.append(Location(**location))
        else:
            self.locations = kwargs.get("locations")



class Municipality:
    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.districtId = kwargs.get("districtId")
        self.districtName = kwargs.get("districtName")
        self.regionId = kwargs.get("regionId")
        self.regionName = kwargs.get("regionName")
        self.isActive = kwargs.get("isActive")
        self.id = kwargs.get("id")



from loguru import logger

from komora_syncer.connections.NetboxConnection import NetboxConnection


class NetboxCache:
    nb_client = NetboxConnection.get_connection()
    tenant_cache = {}
    provider_cache = {}
    device_cache = {}
    site_cache = {}
    region_cache = {}

    def __init__(self):
        raise TypeError("This is a static class")

    @classmethod
    def init_tenant_cache(cls, force=False):
        if cls.tenant_cache and not force:
            return cls.tenant_cache

        try:
            tenants = cls.nb_client.tenancy.tenants.all()
            for tenant in tenants:
                if tenant.custom_fields.get("komora_id"):
                    cls.tenant_cache[tenant.custom_fields.get("komora_id")] = tenant
        except Exception:
            logger.exception("Unable to initialize tenant cache")
            raise

    @classmethod
    def init_provider_cache(cls, force=False):
        if cls.provider_cache and not force:
            return cls.provider_cache

        try:
            providers = cls.nb_client.circuits.providers.all()
            for provider in providers:
                if provider.custom_fields.get("komora_id"):
                    cls.provider_cache[provider.custom_fields.get("komora_id")] = provider
        except Exception:
            logger.exception("Unable to initialize tenant cache")
            raise

    @classmethod
    def init_device_cache(cls, force=False):
        if cls.device_cache and not force:
            return cls.device_cache

        try:
            devices = cls.nb_client.dcim.devices.all()
            for device in devices:
                if device.custom_fields.get("komora_id"):
                    cls.device_cache[device.custom_fields.get("komora_id")] = device
        except Exception:
            logger.exception("Unable to initialize device cache")
            raise

    @classmethod
    def init_site_cache(cls, force=False):
        if cls.site_cache and not force:
            return cls.site_cache

        try:
            sites = cls.nb_client.dcim.sites.all()
            for site in sites:
                if site.custom_fields.get("komora_id"):
                    cls.site_cache[site.custom_fields.get("komora_id")] = site
        except Exception:
            logger.exception("Unable to initialize site cache")
            raise

    @classmethod
    def init_region_cache(cls, force=False):
        if cls.region_cache and not force:
            return cls.region_cache
        try:
            regions = cls.nb_client.dcim.regions.all()
            for region in regions:
                if region.custom_fields.get("komora_id"):
                    cls.region_cache[region.custom_fields.get("komora_id")] = region
        except Exception:
            logger.exception("Unable to initialize region cache")
            raise

    @classmethod
    def get_tenant_cache(cls):
        if not cls.tenant_cache:
            cls.init_tenant_cache()
        return cls.tenant_cache

    @classmethod
    def get_provider_cache(cls):
        if not cls.provider_cache:
            cls.init_provider_cache()
        return cls.provider_cache

    @classmethod
    def get_device_cache(cls):
        if not cls.device_cache:
            cls.init_device_cache()
        return cls.device_cache

    @classmethod
    def get_site_cache(cls):
        if not cls.site_cache:
            cls.init_site_cache()
        return cls.site_cache

    @classmethod
    def get_region_cache(cls):
        if not cls.region_cache:
            cls.init_region_cache()
        return cls.region_cache



class Address:
    def __init__(self, **kwargs):
        self.municipalityId = kwargs.get("municipalityId")
        self.municipalityName = kwargs.get("municipalityName")
        self.districtId = kwargs.get("districtId")
        self.districtName = kwargs.get("districtName")
        self.regionId = kwargs.get("regionId")
        self.regionName = kwargs.get("regionName")
        self.houseNumber = kwargs.get("houseNumber")
        self.referenceNumber = kwargs.get("referenceNumber")
        self.zip = kwargs.get("zip")
        self.streetName = kwargs.get("streetName")
        self.streetId = kwargs.get("streetId")
        self.text = kwargs.get("text")
        self.isActive = kwargs.get("isActive")
        self.id = kwargs.get("id")



import time
import urllib.parse
from base64 import b64encode
from hashlib import sha256
from hmac import new

import requests
from loguru import logger

from komora_syncer.config import get_config

# Log Warnings and higher errors from imported modules
# logging.getLogger(requests.__name__).setLevel(logging.WARNING)
# logging.getLogger('urllib3').setLevel(logging.WARNING)


class KomoraConnection:
    """
    A class for establishing and managing a connection to a Komora instance.
    """

    @staticmethod
    def checkout_url(model, params=""):
        """
        Constructs a signed URL for accessing a Komora API endpoint.

        Args:
            model (str): The name of the API endpoint.
            params (str): Optional query parameters to include in the URL.

        Returns:
            str: The signed URL.
        """
        private_key, sign_app = get_config()["komora"]["PRIVATE_KEY"], get_config()["komora"]["SIGN_APP"]
        komora_url = get_config()["komora"]["KOMORA_API_URL"]

        base_url = f"{komora_url}/{model}"
        signDate = str(int(time.time()))

        if params:
            params = f"?{params}&signApp={sign_app}&signDate={signDate}"
        else:
            params = f"?signApp={sign_app}&signDate={signDate}"

        digest = new(bytes(private_key, "utf-8"), msg=str.encode(params), digestmod=sha256).digest()
        signature = b64encode(digest).decode()
        parsed_signature = urllib.parse.quote_plus(signature)

        result_url = f"{base_url}{params}&signHash={parsed_signature}"
        return result_url

    @staticmethod
    def __get_page(api, page, page_size, filter_params=""):
        """
        Retrieves a page of records from a Komora API endpoint.

        Args:
            api (str): The name of the API endpoint.
            page (int): The page number to retrieve.
            page_size (int): The number of records per page.
            filter_params (str): Optional filter parameters to include in the request.

        Returns:
            tuple: A tuple containing the list of records and the total number of records.
        """
        params = f"Page={page}&PageSize={page_size}"
        if filter_params:
            params = params + "&" + filter_params

        url = KomoraConnection.checkout_url(api, params)

        try:
            response = requests.get(url)
            response.raise_for_status()
        except Exception as e:
            logger.exception("Unable to get page")
            logger.debug(f"URL {url}")
            raise e
        else:
            response = response.json()
            result = response.get("data", [])

            # Check number of all records
            total = response.get("total", 0)

            return result, int(total)

    @staticmethod
    def get_records(api, filter_params=""):
        """
        Retrieves all records from a Komora API endpoint.

        Args:
            api (str): The name of the API endpoint.
            filter_params (str): Optional filter parameters to include in the request.

        Returns:
            list: A list of all records from the API endpoint.
        """
        page_size = 100
        page = 0

        try:
            # load first page
            record_list, total = KomoraConnection.__get_page(api, page, page_size, filter_params=filter_params)

            # if less records are returned than total number of records -> get rest of records
            while len(record_list) < total:
                page += 1
                try:
                    next_data = KomoraConnection.__get_page(api, page, page_size, filter_params=filter_params)[0]
                    record_list.extend(next_data)
                except Exception as e:
                    logger.exception("Unable to get page")
                    logger.exception(e)
                    logger.debug(f"API {api}, page {page}")
                    break

            if len(record_list) == total:
                return record_list
            # TODO: generate proper error
            else:
                raise Exception
        except Exception as e:
            logger.critical("Unable to get records")
            logger.exception(e)
            logger.debug(f"API {api}")



import json
import time

import requests
from loguru import logger

from komora_syncer.config import get_config
from komora_syncer.connections.KomoraConnection import KomoraConnection
from komora_syncer.helpers.utils import build_tenant_name
from komora_syncer.models.komora.contact import Contact
from komora_syncer.models.komora.device import Device
from komora_syncer.models.komora.district import District
from komora_syncer.models.komora.municipality import Municipality
from komora_syncer.models.komora.organization import Organization
from komora_syncer.models.komora.region import Region
from komora_syncer.models.komora.site import Site

# Log Warnings and higher errors from imported modules
# logging.getLogger(requests.__name__).setLevel(logging.WARNING)
# logging.getLogger("urllib3").setLevel(logging.WARNING)


class KomoraApi:
    cache = {
        "sites": [],
        "regions": [],
        "districts": [],
        "municipalities": [],
        "organizations": [],
        "contacts": [],
        "devices": [],
    }

    def __new__(self):
        raise TypeError("This is a static class")

    @staticmethod
    def store_cache(object_type, objects):
        KomoraApi.cache[object_type] = objects

    @staticmethod
    def get_object_instance(object_type, komora_data):
        object_map = {
            "sites": Site,
            "regions": Region,
            "districts": District,
            "municipalities": Municipality,
            "organizations": Organization,
            "organization_suppliers": Organization,
            "devices": Device,
            "contacts": Contact,
        }
        if object_type not in object_map:
            raise ValueError(f"Unknown object type: {object_type}")
        return object_map[object_type](**komora_data)

    @staticmethod
    def check_cache(object_type):
        return KomoraApi.cache.get(object_type, [])

    @staticmethod
    def get_objects(object_type, api, filter_params=""):
        data = []

        try:
            logger.info(f"Polling {object_type} from Komora")
            for obj in KomoraConnection.get_records(api, filter_params=filter_params):
                data.append(KomoraApi.get_object_instance(object_type, obj))
        except Exception:
            logger.exception(f"Unable to poll {object_type} from Komora")
            raise

        return data

    @staticmethod
    def get_sites():
        sites = KomoraApi.check_cache("sites")
        if not sites:
            sites = KomoraApi.get_objects("sites", "Site", "")
            KomoraApi.store_cache("sites", sites)
        return sites

    @staticmethod
    def get_organizations():
        organizations = KomoraApi.check_cache("organizations")
        if not organizations:
            organizations = KomoraApi.get_objects("organizations", "Organization", "")
            # Filter out duplicated CID and Name
            unique_cids_names = set()
            filtered_organizations = []
            for organization in organizations:
                cid_name = build_tenant_name(organization.clientId, organization.name)
                if cid_name not in unique_cids_names:
                    unique_cids_names.add(cid_name)
                    filtered_organizations.append(organization)
                else:
                    logger.critical(f"Duplicated CID and Name: {cid_name}")
            KomoraApi.store_cache("organizations", filtered_organizations)
            organizations = filtered_organizations
        return organizations

    @staticmethod
    def get_supplier_organizations():
        organizations = KomoraApi.check_cache("organization_suppliers")
        if not organizations:
            organizations = KomoraApi.get_objects("organization_suppliers", "Organization", "Filters.isSupplier=true")

            # Filter out duplicated CID and Name
            unique_cids_names = set()
            filtered_organizations = []
            for organization in organizations:
                cid_name = build_tenant_name(organization.clientId, organization.name)
                if cid_name not in unique_cids_names:
                    unique_cids_names.add(cid_name)
                    filtered_organizations.append(organization)
                else:
                    logger.critical(f"Duplicated CID and Name: {cid_name}")
            KomoraApi.store_cache("organization_suppliers", filtered_organizations)
            organizations = filtered_organizations
        return organizations

    @staticmethod
    def get_regions():
        regions = KomoraApi.check_cache("regions")
        if not regions:
            regions = KomoraApi.get_objects("regions", "Ruian/region")
            KomoraApi.store_cache("regions", regions)
        return regions

    @staticmethod
    def get_districts():
        districts = KomoraApi.check_cache("districts")
        if not districts:
            districts = KomoraApi.get_objects("districts", "Ruian/district")
            KomoraApi.store_cache("districts", districts)
        return districts

    @staticmethod
    def get_municipalities():
        municipalities = KomoraApi.check_cache("municipalities")
        if not municipalities:
            municipalities = KomoraApi.get_objects("municipalities", "Ruian/municipality")
            KomoraApi.store_cache("municipalities", municipalities)
        return municipalities

    @staticmethod
    def get_devices():
        devices = KomoraApi.check_cache("devices")
        if not devices:
            filter_params = "Filters.IsActive=true"
            devices = KomoraApi.get_objects("devices", "Device", filter_params)
            KomoraApi.store_cache("devices", devices)
        return devices

    @staticmethod
    def post_devices(nb_devices):
        # Save nb_devices to file
        time_str = time.strftime("%Y%m%d-%H%M%S")
        filename = f"{time_str}_devices.json"
        filepath = get_config()["common"]["DEVICES_DATA_PATH"]
        file = filepath + filename

        with open(file, "w") as outfile:
            json.dump(nb_devices, outfile)

        url = KomoraConnection.checkout_url("ServiceRecord/SendDeviceData")

        try:
            logger.info("Posting devices to Komora")
            requests.post(url, files={"dataFile": open(file, "rb")})
        except requests.exceptions.RequestException:
            logger.exception("Unable to post devices to Komora")
            raise

    @staticmethod
    def get_contacts():
        contacts = KomoraApi.check_cache("contacts")
        if not contacts:
            contacts = KomoraApi.get_objects("contacts", "Contact")
            KomoraApi.store_cache("contacts", contacts)
        return contacts


from komora_syncer.helpers.utils import clean_organization_name


class Organization:
    def __init__(self, **kwargs):
        self.id = kwargs.get("id")
        self.abraCustomerId = kwargs.get("abraCustomerId")
        self.addres = kwargs.get("address")
        self.hasActiveContract = kwargs.get("hasActiveContract")
        self.icoNumber = kwargs.get("icoNumber")
        self.isActual = kwargs.get("isActual")
        self.isCustomer = kwargs.get("isCustomer")
        self.name = clean_organization_name(kwargs.get("name"))
        self.abraCode = kwargs.get("abraCode")
        self.xUcastnikId = kwargs.get("xUcastnikId")
        self.clientId = kwargs.get("clientId")
        self.updatedAt = kwargs.get("updatedAt")
        self.updatedBy = kwargs.get("updatedBy")
        self.isMember = kwargs.get("isMember")
        self.isCanceledMemberPart = kwargs.get("isCanceledMemberPart")
        self.masterOrganizationId = kwargs.get("masterOrganizationId")
        self.masterOrganizationAbraId = kwargs.get("masterOrganizationAbraId")
        self.masterOrganizationName = kwargs.get("masterOrganizationName")
        self.entityId = kwargs.get("entityId")
        self.isVip = kwargs.get("isVip")
        self.vipNote = kwargs.get("vipNote")
        self.organizationTypeId = kwargs.get("organizationTypeId")
        self.organizationTypeName = kwargs.get("organizationTypeName")
        self.isSupplier = kwargs.get("isSupplier")
        self.keywords = kwargs.get("keywords")



import pynetbox
from loguru import logger

from komora_syncer.config import get_config


class NetboxConnection:
    netbox_url = get_config()["netbox"]["NETBOX_INSTANCE_URL"]
    api_token = get_config()["netbox"]["NETBOX_API_TOKEN"]
    connection = None

    def __new__(cls):
        raise TypeError("This is a static class")

    @staticmethod
    def open():
        """
        Opens a connection to the NetBox instance using the configured URL and API token.
        """
        try:
            NetboxConnection.connection = pynetbox.api(
                NetboxConnection.netbox_url, token=NetboxConnection.api_token, threading=True
            )
        except Exception:
            logger.exception("Could not establish connection to NetBox instance")
            raise

    @staticmethod
    def get_connection():
        """
        Returns the connection to the NetBox instance, opening it if necessary.
        """
        if NetboxConnection.connection is None:
            try:
                NetboxConnection.open()
            except Exception:
                logger.exception("Could not open connection to NetBox instance")
                return None

        return NetboxConnection.connection



from komora_syncer.models.komora.address import Address
from komora_syncer.models.komora.location import Location
from komora_syncer.models.komora.site_contact import SiteContact


class Site:
    def __init__(self, **kwargs):
        self.name = (kwargs.get("name") or "").strip()
        self.fullName = (kwargs.get("fullName") or "").strip()
        self.description = (kwargs.get("description") or "").strip()
        self.facility = kwargs.get("facility")
        self.latitude = kwargs.get("latitude")
        self.longitude = kwargs.get("longitude")
        self.code = kwargs.get("code")
        self.typeId = kwargs.get("typeId")
        self.typeName = kwargs.get("typeName")
        self.isActive = kwargs.get("isActive")
        self.id = kwargs.get("id")

        self.organizationId = kwargs.get("organizationId")
        self.organizationName = kwargs.get("organizationName")

        if kwargs.get("address", None):
            self.address = Address(**kwargs.get("address"))
        else:
            self.address = kwargs.get("address")

        if kwargs.get("locations", None):
            self.locations = []
            for location in kwargs.get("locations"):
                self.locations.append(Location(**location))
        else:
            self.locations = kwargs.get("locations")

        if kwargs.get("contacts", None):
            self.contacts = []
            for contact in kwargs.get("contacts"):
                self.contacts.append(SiteContact(**contact))
        else:
            self.contacts = kwargs.get("contacts")

        self.flatten_locations = flatten_locations(self.locations) if self.locations else []

        self.regionId = kwargs.get("regionId")
        self.regionName = kwargs.get("regionName")


def flatten_locations(nested_locations):
    result = []

    def flat(nested_locations):
        for location in nested_locations:
            result.append(location)
            flat(location.locations)
        return result

    flat(nested_locations)
    return result



from datetime import datetime
from typing import List


class ContactType:
    contact_id: int
    contract_id: int
    contact_type_id: int
    contact_type_name: str
    id: int

    def __init__(self, **kwargs) -> None:
        self.contact_id = kwargs.get("contact_id")
        self.contract_id = kwargs.get("contract_id")
        self.contact_type_id = kwargs.get("contact_type_id")
        self.contact_type_name = kwargs.get("contact_type_name")
        self.id = kwargs.get("id")


class Email:
    email_address: str

    def __init__(self, **kwargs) -> None:
        self.email_address = kwargs.get("email_address")


class Organization:
    organization_name: str
    organization_id: int
    isActive: bool

    def __init__(self, **kwargs) -> None:
        self.organization_name = kwargs.get("organization_name")
        self.organization_id = kwargs.get("organization_id")
        self.isActive = kwargs.get("isActive")


class Tag:
    contact_id: int
    tag_id: int
    tag_name: str
    note: str
    id: int

    def __init__(self, **kwargs) -> None:
        self.contact_id = kwargs.get("contact_id")
        self.tag_id = kwargs.get("tag_id")
        self.tag_name = kwargs.get("tag_name")
        self.note = kwargs.get("note")
        self.id = kwargs.get("id")


class Contact:
    firstName: str
    surname: str
    title: str
    suffix: str
    contactTypes: List[ContactType]
    phoneNumber: str
    isActive: bool
    address: str
    phoneNumber2: str
    street: str
    city: str
    postCode: str
    validFrom: datetime
    validTo: datetime
    abraContactId: str
    organizations: List[Organization]
    updatedAt: datetime
    updatedBy: str
    fullName: str
    isEditable: bool
    lastContactedNote: str
    lastContactedBy: str
    lastContactedAt: datetime
    emails: List[Email]
    tags: List[Tag]
    id: int

    def __init__(self, **kwargs) -> None:
        self.firstName = kwargs.get("firstName")
        self.surname = kwargs.get("surname")
        self.title = kwargs.get("title")
        self.suffix = kwargs.get("suffix")
        self.contactTypes = kwargs.get("contactTypes")
        self.phoneNumber = kwargs.get("phoneNumber")
        self.isActive = kwargs.get("isActive")
        self.address = kwargs.get("address")
        self.phoneNumber2 = kwargs.get("phoneNumber2")
        self.street = kwargs.get("street")
        self.city = kwargs.get("city")
        self.postCode = kwargs.get("postCode")
        self.validFrom = kwargs.get("validFrom")
        self.validTo = kwargs.get("validTo")
        self.abraContactId = kwargs.get("abraContactId")
        self.organizations = kwargs.get("organizations")
        self.updatedAt = kwargs.get("updatedAt")
        self.updatedBy = kwargs.get("updatedBy")
        self.fullName = kwargs.get("fullName")
        self.isEditable = kwargs.get("isEditable")
        self.lastContactedNote = kwargs.get("lastContactedNote")
        self.lastContactedBy = kwargs.get("lastContactedBy")
        self.lastContactedAt = kwargs.get("lastContactedAt")
        self.emails = kwargs.get("emails")
        self.tags = kwargs.get("tags")
        self.id = kwargs.get("id")


"""
 "data": [
    {
      "firstName": "string",
      "surname": "string",
      "title": "string",
      "suffix": "string",
      "contactTypes": [
        {
          "contactId": 0,
          "contractId": 0,
          "contactTypeId": 0,
          "contactTypeName": "string",
          "id": 0
        }
      ],
      "phoneNumber": "string",
      "isActive": true,
      "address": "string",
      "phoneNumber2": "string",
      "street": "string",
      "city": "string",
      "postCode": "string",
      "validFrom": "2021-11-08T08:49:44.972Z",
      "validTo": "2021-11-08T08:49:44.972Z",
      "abraContactId": "string",
      "organizations": [
        {
          "organizationName": "string",
          "organizationId": 0,
          "isActive": true
        }
      ],
      "updatedAt": "2021-11-08T08:49:44.972Z",
      "updatedBy": "string",
      "fullName": "string",
      "isEditable": true,
      "lastContactedNote": "string",
      "lastContactedBy": "string",
      "lastContactedAt": "2021-11-08T08:49:44.972Z",
      "emails": [
        {
          "emailAddress": "string"
        }
      ],
      "tags": [
        {
          "contactId": 0,
          "tagId": 0,
          "tagName": "string",
          "note": "string",
          "id": 0
        }
      ],
      "id": 0
    }
  ]
"""



import traceback

import click
from loguru import logger

from komora_syncer.config import configure_logger
from komora_syncer.connections.NetboxConnection import NetboxConnection
from komora_syncer.helpers.utils import check_required_custom_fields
from komora_syncer.processors.Synchronizer import Synchronizer

SYNC_OPTIONS = ["all", "organizations", "regions", "sites", "devices"]

# Initialize logging configuration
configure_logger()


@click.group()
def cli():
    pass


@click.command()
@click.option(
    "--sync",
    "-s",
    type=click.Choice(SYNC_OPTIONS),
    multiple=True,
    default=["all"],
    help="what should be synced",
)
@click.option("--site", "-t", type=click.STRING, help="Name of site to deploy")
@click.option("--skip-custom-fields-check", is_flag=True, help="Skip checking for required custom fields in NetBox")
def synchronize(sync, site, skip_custom_fields_check):
    """
    Synchronize data between Netbox and Komora
    """
    # Check if all required custom fields exist in NetBox before starting synchronization
    if not skip_custom_fields_check:
        logger.info("Checking required custom fields in NetBox...")
        nb_connection = NetboxConnection.get_connection()
        check_required_custom_fields(nb_connection)
    else:
        logger.info("Skipping custom fields check as requested")

    synchronizer = Synchronizer()

    if site and "sites" not in sync:
        logger.critical("Site option '--site' is only available with sites 'synchronize --sync sites' synchronization")
        logger.critical("Exiting")
        return

    if site and "sites" in sync:
        try:
            # Syncs Sites / Site, Location
            synchronizer.sync_sites(site_name=site)
            return
        except Exception as e:
            logger.error(f"Unable to synchronize site '{site}' - {e}")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if "all" in sync or "organizations" in sync:
        try:
            # Syncs Organizations / Tenants
            synchronizer.sync_organizations()
        except Exception as e:
            logger.error("Unable to synchronize organizations")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if "all" in sync or "organizations" in sync:
        try:
            # Syncs Organizations / Suppliers
            synchronizer.sync_organization_suppliers()
        except Exception as e:
            logger.error("Unable to synchronize organizations (supplier)")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if "all" in sync or "regions" in sync:
        try:
            # Syncs Regions, Disctricts and Municipalities / Regions
            synchronizer.sync_regions()
        except Exception as e:
            logger.error("Unable to synchronize regions")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if "all" in sync or "sites" in sync:
        try:
            # Syncs Sites / Site, Location
            synchronizer.sync_sites()
        except Exception as e:
            logger.error("Unable to synchronize sites")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if "all" in sync or "devices" in sync:
        try:
            synchronizer.sync_devices()
        except Exception as e:
            logger.error("Unable to synchronize devices")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return


cli.add_command(synchronize)


if __name__ == "__main__":
    # Display CLI
    cli()



from lat_lon_parser import parse
from loguru import logger


def batch_query(method, id_list, batch_size=50, id_param="id__in", **kwargs):
    """
    Execute a query in batches to avoid Request-URI Too Long errors.

    Args:
        method: The pynetbox query method to call (e.g., nb.dcim.interfaces.filter)
        id_list: List of IDs to split into batches
        batch_size: Maximum number of IDs per batch
        id_param: Parameter name for the ID filter (default: 'id__in')
        **kwargs: Additional filter parameters to pass to the query method

    Returns:
        List of results from all batches combined
    """
    results = []
    total_batches = (len(id_list) + batch_size - 1) // batch_size

    for i in range(0, len(id_list), batch_size):
        batch = id_list[i : i + batch_size]
        logger.debug(f"Processing batch {(i // batch_size) + 1}/{total_batches} with {len(batch)} items")

        # Create params dict with the batch of IDs
        batch_params = {id_param: batch}
        # Add any additional filter parameters
        batch_params.update(kwargs)

        # Execute the query and extend the results
        batch_results = list(method(**batch_params))
        results.extend(batch_results)

    logger.debug(f"Query completed with {len(results)} total results")
    return results


def build_tenant_name(client_id, name):
    # Clean the name to remove quotation marks
    cleaned_name = clean_organization_name(name) if name else name
    return f"{client_id} - {cleaned_name}"[0:100].strip()


def clean_organization_name(name):
    """
    Remove double quotation marks from organization names.
    Removes only double (") quotes from the organization name.

    Args:
        name (str): The organization name to clean

    Returns:
        str: The cleaned organization name
    """
    if not name:
        return name

    # Remove only double quotes from the name
    cleaned_name = name.replace('"', "")
    return cleaned_name.strip()


def sanitize_lat_lon(lat_or_lon: str) -> float:
    """
    Gets latitude or longitude string and returns float with 6 decimal places
    This format is required in Netbox
    """
    if not lat_or_lon:
        return None

    try:
        result = parse(lat_or_lon)
    except ValueError:
        return None

    if result:
        return round(result, 6)
    else:
        return None


def check_required_custom_fields(nb_connection, exit_on_error=True):
    """
    Check if all required custom fields exist in NetBox.
    Exits the program with an error message if any custom fields are missing.

    Args:
        nb_connection: NetBox API connection instance
        exit_on_error: Whether to exit the program on error (default: True)

    Returns:
        bool: True if all fields exist or check was skipped, False if there are missing fields
    """
    # Define all required custom fields organized by NetBox model
    required_custom_fields = {
        "tenancy.tenant": [
            "client_id",
            "komora_id",
            "client_name",
            "komora_is_customer",
            "komora_is_supplier",
            "komora_is_vip",
            "komora_is_actual",
            "komora_is_member",
            "komora_vip_note",
        ],
        "circuits.provider": ["client_id", "client_name", "komora_id", "tenant"],
        "dcim.site": ["komora_id", "komora_url", "code", "type"],
        "dcim.region": ["komora_id", "region_type"],
        "dcim.location": ["komora_id"],
        "dcim.device": ["komora_id", "komora_url"],
    }

    missing_fields = []
    connectivity_errors = []

    for model_name, field_names in required_custom_fields.items():
        try:
            # Get the custom fields for this model
            app_label, model = model_name.split(".")
            custom_fields_response = nb_connection.extras.custom_fields.filter(content_types=model_name)

            # Extract existing field names
            existing_fields = {cf.name for cf in custom_fields_response}

            # Check for missing fields
            for field_name in field_names:
                if field_name not in existing_fields:
                    missing_fields.append(f"{model_name}.{field_name}")

        except Exception as e:
            error_msg = str(e)
            logger.exception(f"Unable to check custom fields for model {model_name}")

            # Check if this is a connectivity error (503, 502, 500, connection errors)
            if any(code in error_msg for code in ["503", "502", "500", "Connection", "connection"]):
                connectivity_errors.append(f"{model_name}.* (connectivity issue: {error_msg})")
            else:
                missing_fields.append(f"{model_name}.* (unable to check: {error_msg})")

    # Handle connectivity errors differently
    if connectivity_errors and not missing_fields:
        logger.warning("Unable to check custom fields due to connectivity issues with NetBox:")
        for field in sorted(connectivity_errors):
            logger.warning(f"  - {field}")
        logger.warning("NetBox appears to be unavailable. Proceeding with synchronization...")
        logger.warning("Please ensure NetBox is accessible and custom fields are properly configured.")
        return True

    if missing_fields:
        logger.critical("Missing required custom fields in NetBox:")
        for field in sorted(missing_fields):
            logger.critical(f"  - {field}")
        logger.critical("\nPlease create these custom fields in NetBox before running the synchronizer.")

        if exit_on_error:
            exit(1)
        return False

    logger.info("All required custom fields are present in NetBox")
    return True



from loguru import logger
from slugify import slugify

from komora_syncer.models.netbox.NetboxBase import NetboxBase
from komora_syncer.models.netbox.NetboxCache import NetboxCache


class NbRegion(NetboxBase):
    REGION_TYPES = ["region", "district", "municipality"]

    def __init__(self, komora_obj=None, region_type=None):
        super().__init__()
        self.api_object = None

        self.name = komora_obj.name
        self.slug = slugify(self.name)
        self.komora_id = komora_obj.id
        self.region_type = region_type

        self.parent = None
        self._find_parent(komora_obj, region_type)

    def _find_parent(self, komora_obj, region_type):
        if self.slug == "zahranicni-kraj":
            return

        parent_name, parent_type = self._get_parent_info(komora_obj, region_type)
        if not parent_name:
            return

        try:
            self.parent = self.nb.dcim.regions.get(
                name=parent_name,
                cf_region_type=parent_type,
                cf_komora_id__n=self.komora_id,
            )
        except Exception as e:
            logger.critical(f"Parent region {parent_name} does not exist")
            logger.exception(e)

    def _get_parent_info(self, komora_obj, region_type):
        if region_type == "region":
            return "CZ", "state"
        elif region_type == "district":
            return komora_obj.regionName, "region"
        elif region_type == "municipality":
            return komora_obj.districtName, "district"
        else:
            raise ValueError(f"Invalid region type: {region_type}")

    def find_or_create(self):
        self.find()
        if not self.api_object:
            self.create()

        return self.api_object

    def find(self):
        if self.komora_id:
            try:
                if self.komora_id in NetboxCache.get_region_cache():
                    netbox_region = NetboxCache.get_region_cache()[self.komora_id]
                else:
                    netbox_region = self.nb.dcim.regions.get(cf_komora_id=self.komora_id)

                if netbox_region:
                    self.api_object = netbox_region
                    return self.api_object
            except Exception as e:
                logger.exception(f"Unable to find region by komora_id: {self.komora_id}")
                logger.debug(e)

    def create(self):
        try:
            netbox_region_param = self.params
            netbox_region = self.nb.dcim.regions.create(netbox_region_param)

            logger.info("Region: %s created successfully", self.name)
            self.api_object = netbox_region
        except Exception as e:
            logger.exception("Unable to create netbox site: %s", self.name)
            raise e

        return self.api_object

    def update(self, nb_region):
        try:
            if nb_region.update(self.params):
                self.api_object = nb_region
                logger.info(f"Region: {self.name} updated successfully")
        except Exception as e:
            logger.exception(f"Unable to update region {self.name}")
            raise e

    def synchronize(self):
        region = self.find()

        if region:
            self.update(region)
        else:
            self.create()

    @property
    def params(self):
        params = {"name": self.name, "slug": self.slug}

        if self.parent:
            params["parent"] = self.parent.id

        if self.komora_id:
            params.setdefault("custom_fields", {})["komora_id"] = self.komora_id

        if self.region_type:
            params.setdefault("custom_fields", {})["region_type"] = self.region_type

        return params



from loguru import logger

from komora_syncer.connections.KomoraApi import KomoraApi
from komora_syncer.connections.NetboxConnection import NetboxConnection
from komora_syncer.models.netbox.NbDevice import NbDevice
from komora_syncer.models.netbox.NbProvider import NbProvider
from komora_syncer.models.netbox.NbRegion import NbRegion
from komora_syncer.models.netbox.NbSite import NbSite
from komora_syncer.models.netbox.NbTenant import NbTenant


class Synchronizer:
    """
    A class for synchronizing data between Komora and NetBox.
    """

    def __init__(self):
        self.sites = []
        self.regions_all = []
        self.districts_all = []
        self.municipalities_all = []
        self.regions_assigned = []
        self.districts_assigned = []
        self.municipalities_assigned = []

    def __prepare_all_regions(self):
        """
        Prepares the lists of all regions, municipalities, districts, and sites from Komora API.
        """
        if not self.regions_all:
            self.regions_all = KomoraApi.get_regions()
        if not self.municipalities_all:
            self.municipalities_all = KomoraApi.get_municipalities()
        if not self.districts_all:
            self.districts_all = KomoraApi.get_districts()
        if not self.sites:
            self.sites = KomoraApi.get_sites()

    def __prepare_assigned_regions(self):
        """
        Prepares the lists of assigned regions, municipalities, districts, and sites from Komora API.
        """
        if not self.regions_all or not self.municipalities_all or not self.districts_all or not self.sites:
            self.__prepare_all_regions()

        municipalities_on_sites = [
            [site.address.municipalityName, site.address.districtName] for site in self.sites if site.address
        ]
        self.municipalities_assigned = [
            muni for muni in self.municipalities_all if [muni.name, muni.districtName] in municipalities_on_sites
        ]

        districts_on_assigned_municipalities = set(
            municipality.districtName for municipality in self.municipalities_assigned
        )
        self.districts_assigned = [
            dis for dis in self.districts_all if dis.name in districts_on_assigned_municipalities
        ]

        regions_on_assigned_districts = set(district.regionName for district in self.districts_assigned)
        self.regions_assigned = [reg for reg in self.regions_all if reg.name in regions_on_assigned_districts]

        sites_without_address_with_region = [
            [site.regionId, site.regionName] for site in self.sites if not site.address and site.regionId
        ]
        self.regions_assigned.extend(
            [reg for reg in self.regions_all if [reg.id, reg.name] in sites_without_address_with_region]
        )

    def sync_regions(self):
        """
        Synchronizes regions, districts, and municipalities from Komora API to NetBox.
        """
        if (
            not self.regions_assigned
            or not self.municipalities_assigned
            or not self.districts_assigned
            or not self.sites
        ):
            self.__prepare_assigned_regions()

        logger.info("Synchronizing regions")
        for reg in self.regions_assigned:
            try:
                NbRegion(reg, "region").synchronize()
            except Exception as e:
                logger.exception(f"Failed to synchronize region {reg.name}: {e}")
                raise e

        for dis in self.districts_assigned:
            try:
                NbRegion(dis, "district").synchronize()
            except Exception as e:
                logger.exception(f"Failed to synchronize district {dis.name}: {e}")
                raise e

        for muni in self.municipalities_assigned:
            try:
                NbRegion(muni, "municipality").synchronize()
            except Exception as e:
                logger.exception(f"Failed to synchronize municipality {muni.name}: {e}")
                raise e

    def sync_sites(self, site_name=None):
        """
        Synchronizes sites from Komora API to NetBox.
        If site_name is specified, synchronizes only the site with the given name.
        """
        try:
            if not self.sites:
                self.sites = KomoraApi.get_sites()
        except Exception as e:
            raise e

        if site_name:
            logger.info(f"Synchronizing site: {site_name}")
            found_sites = [site for site in self.sites if site.name == site_name]
            if len(found_sites) == 0:
                raise ValueError(f"No object found with name '{site_name}'")
            elif len(found_sites) > 1:
                raise ValueError(f"Multiple objects found with name '{site_name}'")
            else:
                try:
                    NbSite(found_sites[0]).synchronize()
                except Exception as e:
                    logger.exception(f"Failed to synchronize site {site_name}: {e}")
                    raise e
        else:
            logger.info("Synchronizing sites")
            for site in self.sites:
                try:
                    NbSite(site).synchronize()
                except Exception as e:
                    logger.exception(f"Failed to synchronize site {site.name}: {e}")
                    raise e

    def sync_organizations(self):
        """
        Synchronizes organizations from Komora API to NetBox.
        """
        organizations = KomoraApi.get_organizations()

        logger.info("Synchronizing organizations")
        for organization in organizations:
            try:
                NbTenant(organization).synchronize()
            except Exception as e:
                logger.exception(f"Failed to synchronize organization {organization.name}: {e}")
                raise e

    def sync_organization_suppliers(self):
        """
        Synchronizes organizations (suppliers) from Komora API to NetBox Providers.
        """
        organizations = KomoraApi.get_supplier_organizations()

        logger.info("Synchronizing organizations (suppliers)")
        for organization in organizations:
            try:
                NbProvider(organization).synchronize()
            except Exception as e:
                logger.exception(f"Failed to synchronize organizations (suppliers) {organization.name}: {e}")
                raise e

    def sync_devices(self):
        """
        Synchronizes devices from Komora API to NetBox, and unsets links to Komora for devices not found in Komora API.
        """
        try:
            logger.info("Posting devices")
            post_devices = NbDevice.get_nb_devices_data()
            KomoraApi.post_devices(post_devices)
        except Exception as e:
            raise e

        try:
            devices = KomoraApi.get_devices()
        except Exception as e:
            raise e

        logger.info("Synchronizing devices")
        for dev in devices:
            try:
                NbDevice(dev).synchronize()
            except Exception as e:
                logger.exception(f"Failed to synchronize device {dev.name}: {e}")
                raise e

        komora_ids = [dev.id for dev in devices]
        devices_not_in_komora = NetboxConnection.get_connection().dcim.devices.filter(cf_komora_id__n=komora_ids)

        devices_to_unset_links = [dev for dev in devices_not_in_komora]
        for dev in devices_to_unset_links:
            dev.custom_fields["komora_id"] = None
            dev.custom_fields["komora_url"] = None

        try:
            if NetboxConnection.get_connection().dcim.devices.update(devices_to_unset_links):
                logger.info("Devices links to Komora successfully unset")
            else:
                logger.info("No device found for unset links to Komora")
        except Exception as e:
            raise e



from loguru import logger
from slugify import slugify

from komora_syncer.models.netbox.NetboxBase import NetboxBase


class NbLocation(NetboxBase):
    def __init__(self, location_komora, parent_location_komora, nb_site):
        super().__init__()
        self.name = str(location_komora.name or "").strip()
        self.slug = slugify(self.name)
        self.description = str(location_komora.description or "").strip()
        self.komora_id = location_komora.id
        self.api_object = None
        self.site = nb_site
        self.parent = None

        if parent_location_komora:
            nb_parent_location = NbLocation(parent_location_komora, None, nb_site)
            if nb_parent_location.find():
                self.parent = nb_parent_location.api_object
            else:
                logger.critical(f"Parent object exists: {parent_location_komora.name}, but not found in Netbox")

    def find_or_create(self):
        self.find()
        if not self.api_object:
            self.create()
        return self.api_object

    def find(self):
        try:
            netbox_location = self.nb.dcim.locations.get(cf_komora_id=self.komora_id)
            if netbox_location:
                self.api_object = netbox_location
        except Exception as e:
            logger.exception(f"Unable to get location {self.name} {self.site.name} from Netbox")
            raise e
        return self.api_object

    def create(self):
        try:
            netbox_location = self.nb.dcim.locations.create(self.params)

            if netbox_location:
                logger.info("Location: %s created successfully", self.name)
                self.api_object = netbox_location
            else:
                logger.critical(f"Unable to create location {self.name}, site: {self.site_id}")
        except Exception as e:
            logger.exception("Unable to create Netbox location: %s", self.name)
            raise e
        return self.api_object

    def update(self, nb_location):
        try:
            if nb_location.update(self.params):
                self.api_object = nb_location
                logger.info(f"Location: {self.name} updated successfully")
        except Exception as e:
            logger.exception(f"Unable to update location {self.name} in Netbox")
            raise e

    def synchronize(self):
        location = self.find()
        if location:
            self.update(location)
        else:
            self.create()

    @property
    def params(self):
        params = {
            "name": self.name,
            "slug": self.slug,
            "description": self.description,
            "site": self.site.id,
        }
        params.update({"parent": self.parent.id} if self.parent else {})
        params.update({"custom_fields": {"komora_id": self.komora_id}} if self.komora_id is not None else {})
        return params



"""Package configuration."""

from setuptools import find_packages, setup
from komora_syncer import __appname__

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

requirements = [
    "click==8.1.7",
    "pynetbox==7.4.0",
    "python-slugify==8.0.4",
    "appdirs==1.4.4",
    "PyYAML==6.0.2",
    "lat-lon-parser",
    "loguru",
]

setup_requirements = [
    "pytest-runner",
]
test_requirements = ["pytest", "pytest-cov", "pytest-mock", "pynxos", "ipdb", "ruff"]

setup(
    name=__appname__,
    author="Jan Krupa",
    author_email="jan.krupa@cesnet.cz",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "Operating System :: Unix",
    ],
    description="Synchronize data between Komora and Netbox applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    include_package_data=True,
    keywords="netbox,komora",
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://gitlab.cesnet.cz/701/done/netbox_komora_syncer",
    version="5.7",
    zip_safe=False,
    python_requires=">=3.8, <4",
    entry_points={
        "console_scripts": [
            "komora_syncer=komora_syncer.__main__:cli",
        ]
    },
)



from loguru import logger
from slugify import slugify

from komora_syncer.config import get_config
from komora_syncer.helpers.utils import sanitize_lat_lon
from komora_syncer.models.netbox.NbLocation import NbLocation
from komora_syncer.models.netbox.NetboxBase import NetboxBase
from komora_syncer.models.netbox.NetboxCache import NetboxCache


class NbSite(NetboxBase):
    def __init__(self, komora_obj):
        super().__init__()
        # Locations are already flatten via Komora site obj
        self.site_locations = komora_obj.flatten_locations

        self.name = komora_obj.name
        self.slug = slugify(self.name)

        self.description = komora_obj.fullName
        self.comments = komora_obj.description

        self.physical_address = komora_obj.address.text if komora_obj.address else ""
        self.shipping_address = ""
        self.latitude = sanitize_lat_lon(komora_obj.latitude)
        self.longitude = sanitize_lat_lon(komora_obj.longitude)

        self.komora_id = komora_obj.id
        self.komora_url = f"{get_config()['komora']['KOMORA_URL']}/app/site/{self.komora_id}"

        # site id
        self.code = komora_obj.code
        # Types: Připojný bod, virtuální bod, etc
        self.type_name = komora_obj.typeName
        self.type_id = komora_obj.typeId

        self.api_object = None
        self.region = None
        self.tenant = None

        try:
            if komora_obj.address:
                municipality_komora_id = komora_obj.address.municipalityId

                if municipality_komora_id:
                    self.region = self.nb.dcim.regions.get(cf_komora_id=municipality_komora_id)
            elif komora_obj.regionId:
                self.region = self.nb.dcim.regions.get(cf_komora_id=komora_obj.regionId)
        except Exception as e:
            logger.critical(f"Unable to find region {komora_obj.address} for site {self.name}")
            logger.exception(e)

        try:
            komora_org_id = komora_obj.organizationId

            if komora_org_id:
                self.tenant = self.nb.tenancy.tenants.get(cf_komora_id=komora_org_id)
        except Exception as e:
            logger.critical(f"Unable to find tenant {komora_obj.organization} for site {self.name}")
            logger.exception(e)

    def find(self):
        if self.komora_id is None:
            return self.api_object

        site_cache = NetboxCache.get_site_cache()
        netbox_site = (
            site_cache[self.komora_id]
            if self.komora_id in site_cache
            else self.nb.dcim.sites.get(cf_komora_id=self.komora_id)
        )
        if netbox_site:
            self.api_object = netbox_site

        return self.api_object

    def create(self):
        try:
            params = self.params
            netbox_site = self.nb.dcim.sites.create(params)

            logger.info("Site: %s created sucessfully", self.name)
            self.api_object = netbox_site
        except Exception as e:
            logger.exception("Unable to create netbox site: %s", self.name)
            raise e

        return self.api_object

    def update(self):
        try:
            if self.api_object.update(self.params):
                logger.info(f"Site: {self.name} updated successfuly")
        except Exception as e:
            logger.critical(f"Unable to update site {self.name}")
            logger.exception(e)

    def find_or_create(self):
        self.find()
        if self.api_object is None:
            self.create()

        return self.api_object

    def synchronize(self):
        site = self.find()

        if site:
            self.update()
        else:
            self.create()

        # sync location of site
        for location in self.site_locations:
            parent = (
                next(
                    (parent for parent in self.site_locations if parent.id == location.parentId),
                    None,
                )
                if location.parentId
                else None
            )

            nb_location = NbLocation(location, parent, self.api_object)
            nb_location.synchronize()

    @property
    def params(self):
        params = {
            "name": self.name,
            "slug": self.slug,
            "description": self.description,
            "comments": self.comments,
            "physical_address": self.physical_address,
            "shipping_address": self.shipping_address,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "custom_fields": {
                "komora_id": self.komora_id,
                "komora_url": self.komora_url,
                "code": self.code,
                "type": self.type_name,
            },
        }

        if self.region:
            params["region"] = self.region.id

        if self.tenant:
            params["tenant"] = self.tenant.id

        return params



from loguru import logger

from komora_syncer.config import get_config
from komora_syncer.connections.NetboxConnection import NetboxConnection
from komora_syncer.helpers.utils import batch_query
from komora_syncer.models.netbox.NetboxBase import NetboxBase
from komora_syncer.models.netbox.NetboxCache import NetboxCache


class NbDevice(NetboxBase):
    def __init__(self, komora_obj):
        super().__init__()
        self.name = komora_obj.name
        self.komora_id = komora_obj.id
        self.site_name = komora_obj.siteName
        self.komora_url = f"{get_config()['komora']['KOMORA_URL']}/app/device/{self.komora_id}"
        self.api_object = None

    def find(self):
        # 1. lookup Device by KOMORA_ID
        device_cache = NetboxCache.get_device_cache()

        if not self.komora_id:
            return None

        try:
            if self.komora_id in device_cache:
                netbox_device = device_cache[self.komora_id]
            else:
                netbox_device = self.nb.dcim.devices.get(cf_komora_id=self.komora_id)

            if netbox_device is not None:
                self.api_object = netbox_device
                return self.api_object
        except Exception as e:
            logger.exception(f"Unable to get Device by komora_id: {self.komora_id}. Error: {e}")

        # 2. Lookup device by name, if komora id is not presented
        # - log a problem, when the name exists, but komora_id was not found
        try:
            self.site_name = self.site_name if self.site_name else "SKLAD"

            netbox_site = self.nb.dcim.sites.get(name__ie=self.site_name)
            netbox_device = self.nb.dcim.devices.get(name__ie=self.name, site_id=netbox_site.id)

            if netbox_device is not None:
                logger.warning(
                    f"komora_id: {str(self.komora_id)} was not found, but Device {self.name} at site {self.site_name} already exists"
                )
                self.api_object = netbox_device
                return self.api_object
        except Exception as e:
            logger.exception(f"Unable to get Device by name: {self.name}. Error: {e}")

        return self.api_object

    def update(self, nb_device):
        try:
            if nb_device.update(self.params):
                self.api_object = nb_device
                logger.info(f"Device: {self.name} updated successfully")
        except Exception as e:
            logger.exception(f"Unable to update device {self.name}. Error: {e}")

    def synchronize(self):
        device = self.find()

        if device:
            self.update(device)
        else:
            logger.info(f"Device {self.name} - komora_id: {self.komora_id} not found in netbox")

    @property
    def params(self):
        params = {}

        if self.api_object:
            if isinstance(self.api_object.custom_fields, dict):
                params["custom_fields"] = self.api_object.custom_fields
                params["custom_fields"]["komora_id"] = self.komora_id
                params["custom_fields"]["komora_url"] = self.komora_url
        else:
            params["custom_fields"] = {
                "komora_id": self.komora_id,
                "komora_url": self.komora_url,
            }

        return params

    @staticmethod
    def get_info_from_cache(cache, id, include_tenant=False, additional_fields=[]):
        item = cache.get(id)
        if item:
            info = {
                "id": item.id,
                "name": item.name,
                "custom_fields": item.custom_fields,
            }
            for field in additional_fields:
                info[field] = getattr(item, field)
            if include_tenant and item.tenant:
                info["tenant"] = {
                    "id": str(item.tenant.id),
                    "name": item.tenant.name,
                    "custom_fields": item.tenant.custom_fields,
                }
            return info
        return None

    @staticmethod
    def get_nb_devices_data():
        nb = NetboxConnection.get_connection()

        # Fetch necessary data
        filter_device_roles = list(nb.dcim.device_roles.filter(name=["Passive component", "unknown", "Server"]))
        filter_device_role_ids = [device.id for device in filter_device_roles]

        devices = list(nb.dcim.devices.filter(status="active", name__empty=False, role_id__n=filter_device_role_ids))
        device_ids = [device.id for device in devices]
        interfaces = batch_query(nb.dcim.interfaces.filter, device_ids, batch_size=50, id_param="device_id")
        locations = list(nb.dcim.locations.all())
        sites = list(nb.dcim.sites.all())
        unique_tenant_ids = {device.tenant.id for device in devices if device.tenant}
        tenants = list(nb.tenancy.tenants.filter(tenant_id__in=list(unique_tenant_ids)))

        # Create caches
        interface_cache = {iface.device.id: [] for iface in interfaces}
        for iface in interfaces:
            interface_cache[iface.device.id].append(
                {"id": iface.id, "name": iface.name, "description": iface.description}
            )

        location_cache = {loc.id: loc for loc in locations}
        site_cache = {site.id: site for site in sites}
        tenant_cache = {tenant.id: tenant for tenant in tenants}

        # Construct the output data structure
        output_data = {
            "data": {
                "device_list": [
                    {
                        "id": str(device.id),
                        "name": device.name,
                        "primary_ip4": {
                            "id": device.primary_ip4.id,
                            "address": device.primary_ip4.address,
                        }
                        if device.primary_ip4
                        else None,
                        "comments": device.comments,
                        "serial": device.serial,
                        "custom_fields": device.custom_fields,
                        "location": NbDevice.get_info_from_cache(location_cache, device.location.id)
                        if device.location
                        else None,
                        "tenant": NbDevice.get_info_from_cache(tenant_cache, device.tenant.id)
                        if device.tenant
                        else None,
                        "site": NbDevice.get_info_from_cache(site_cache, device.site.id, include_tenant=True)
                        if device.site
                        else None,
                        "interfaces": interface_cache.get(device.id, []),
                    }
                    for device in devices
                ]
            }
        }

        return output_data



from loguru import logger
from slugify import slugify

from komora_syncer.helpers.utils import build_tenant_name, clean_organization_name
from komora_syncer.models.netbox.NetboxBase import NetboxBase
from komora_syncer.models.netbox.NetboxCache import NetboxCache


class NbProvider(NetboxBase):
    # komora_obj -> Organization(isSupplier=True)
    def __init__(self, komora_obj=None):
        super().__init__()
        self.name = build_tenant_name(komora_obj.clientId, komora_obj.name)
        self.slug = slugify(self.name)
        self.description = (
            " / ".join(komora_obj.keywords) if isinstance(komora_obj.keywords, list) else komora_obj.keywords
        )

        tenant = self.nb.tenancy.tenants.get(cf_komora_id=komora_obj.id)
        self.custom_fields = {
            "client_id": komora_obj.clientId,
            "client_name": clean_organization_name(komora_obj.name),
            "komora_id": komora_obj.id,
            "tenant": tenant.id if tenant else None,
        }

        self.api_object = None

    def find_or_create(self):
        self.find()
        if not self.api_object:
            self.create()

        return self.api_object

    def find(self):
        komora_id = self.custom_fields.get("komora_id")
        provider_cache = NetboxCache.get_provider_cache()

        try:
            if komora_id in provider_cache:
                netbox_provider = provider_cache[komora_id]
            else:
                netbox_provider = self.nb.circuits.providers.get(cf_komora_id=komora_id)

            if netbox_provider:
                self.api_object = netbox_provider
        except Exception:
            logger.exception(f"Unable to find provider by komora_id: {komora_id}")
            netbox_provider = None

        if not netbox_provider:
            try:
                netbox_provider = self.nb.circuits.providers.get(name__ie=self.name)

                if netbox_provider:
                    logger.warning(
                        f"komora_id: {str(komora_id)} was not found, but Provider {self.name} already exists"
                    )
                    self.api_object = netbox_provider
            except Exception:
                logger.exception(f"Unable to find provider by name: {self.name}")
                raise

        return self.api_object

    def create(self):
        try:
            netbox_provider = self.nb.circuits.providers.create(self.params)

            logger.info(f"Provider: {self.name} created successfully")
            self.api_object = netbox_provider
        except Exception:
            logger.exception(f"Unable to create netbox provider: {self.name}")
            raise

        return self.api_object

    def update(self, nb_provider):
        try:
            if nb_provider.update(self.params):
                self.api_object = nb_provider
                logger.info(f"Provider: {self.name} updated successfully")
        except Exception as e:
            logger.critical(
                f"Unable to update provider: {self.name} {nb_provider.id}",
            )
            logger.exception(e)

    def synchronize(self):
        provider = self.find()

        if provider:
            self.update(provider)
        else:
            self.create()

    @property
    def params(self):
        return {
            "name": self.name,
            "slug": self.slug,
            "description": self.description,
            "custom_fields": self.custom_fields,
        }
